# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import types

import six

from pycebes.core import functions
from pycebes.core.column import Column
from pycebes.core.expressions import SparkPrimitiveExpression, UnresolvedColumnName
from pycebes.core.sample import DataSample
from pycebes.core.schema import Schema, StorageTypes, VariableTypes
from pycebes.internal.helpers import require
from pycebes.internal.implicits import get_default_session
from pycebes.internal.serializer import to_json


def _parse_column_names(df, *columns):
    """
    Parse the list of column names (as strings) or ``Column`` objects which belong to this Dataframe

    :param df: The ``Dataframe`` to parse the column names
    :type df: Dataframe
    :return: the list of column names, ready to be sent to the server
    """
    result = []
    col_names = set(df.columns)
    for c in columns:
        if isinstance(c, six.text_type):
            require(c in col_names, 'Column not found in this Dataframe: {}'.format(c))
            result.append(c)
        else:
            require(isinstance(c, Column),
                    'Only column names (strings) or `Column` objects are allowed. Got {!r}'.format(c))
            expr = c.expr

            if isinstance(expr, SparkPrimitiveExpression):
                require(expr.df_id == df.id, 'Column from a different Dataframe is not allowed: {!r}'.format(c))
                require(expr.col_name in col_names, 'Column not found in this Dataframe: {}'.format(c))
                result.append(expr.col_name)
            elif isinstance(expr, UnresolvedColumnName):
                require(expr.col_name in col_names, 'Column not found in this Dataframe: {}'.format(c))
                result.append(expr.col_name)
            else:
                raise ValueError('Invalid column: {!r}. Only column objects constructed from a Dataframe is allowed. '
                                 'Use Dataframe.<column_name> or functions.col(<column_name>) to get one'.format(c))
    return result


def _parse_columns(df, *columns):
    """
    Parse the list of column names (as strings) or ``Column`` objects which belong to this Dataframe
    This is similar to :func:`_parse_column_names`, but returns a list of ``Column`` object instead.

    :return: the list of ``Column`` objects
    """
    cols = []
    if len(columns) == 1 and isinstance(columns[0], types.GeneratorType):
        columns = list(columns[0])

    for c in columns:
        if isinstance(c, six.text_type):
            cols.append(df[c])
        else:
            require(isinstance(c, Column), 'Expect a column or a column name, got {!r}'.format(c))
            cols.append(c)
    return cols


@six.python_2_unicode_compatible
class Dataframe(object):
    """
    Representation of a Cebes Dataframe on the client side. All functions in this
    class result in remote call to the Cebes server to perform corresponding actions.

    Users should **NOT** manually construct this class.
    """

    def __init__(self, _id, _schema):
        """Construct a new Dataframe instance.
        Should not be used by end-users.
        """
        self._id = _id
        self._schema = _schema

    """
    Helpers
    """

    @property
    def _client(self):
        """
        Default client, taken from the current default session

        :rtype: pycebes.core.client.Client
        """
        return get_default_session().client

    def _df_command(self, cmd='', **kwargs):
        """
        Helper to send a POST request to server, and parse the result as a Dataframe

        :rtype: Dataframe
        """
        r = self._client.post_and_wait('df/{}'.format(cmd), kwargs)
        return Dataframe.from_json(r)

    @classmethod
    def from_json(cls, js_data):
        """
        Return a ``Dataframe`` instance from its JSON representation

        # Arguments
        js_data (dict): a dict with ``id`` and ``schema``

        # Returns
        Dataframe: the result Dataframe object
        """
        require('id' in js_data and 'schema' in js_data, 'Invalid Dataframe JSON: {!r}'.format(js_data))
        return Dataframe(_id=js_data['id'], _schema=Schema.from_json(js_data['schema']))

    """
    Public properties and Python magics
    """

    @property
    def id(self):
        """
        Return the unique ID of this :class:`Dataframe`
        """
        return self._id

    @property
    def schema(self):
        """
        The Schema of this data frame
        """
        return self._schema

    @property
    def shape(self):
        """
        Return a 2-tuple with number of rows and columns

        # Example

        ```python
            df.shape
                (540, 40)
        ```
        """
        return len(self), len(self.columns)

    @property
    def columns(self):
        """
        Return a list of column names in this ``Dataframe``
        """
        return self.schema.columns

    def __len__(self):
        """
        Number of rows in this ``Dataframe``
        """
        return self._client.post_and_wait('df/count', data={'df': self.id})

    def __repr__(self):
        return '{}(id={!r})'.format(self.__class__.__name__, self.id)

    def __getattr__(self, item):
        if item in self.columns:
            return Column(SparkPrimitiveExpression(self._id, item))
        raise AttributeError('Attribute not found: {!r}'.format(item))

    def __getitem__(self, item):
        try:
            return self.__getattr__(item)
        except AttributeError:

            if isinstance(item, six.text_type):
                # this might seem non-sense at first, but it will help in complicated queries, e.g.
                # df.alias('left').join(...).select(df['left.*'])
                # but this is only allowed in __getitem__(), not __getattr__()
                return functions.col(item)

            raise KeyError('Column not found: {!r}'.format(item))

    def __dir__(self):
        return dir(type(self)) + list(self.__dict__.keys()) + self.columns

    def with_storage_type(self, column, storage_type=StorageTypes.INTEGER):
        """
        Cast the given column of this Dataframe into the given storage type.
        No-op if the new storage type is the same with the current one.

        # Arguments
        column: a string or a Column object
        storage_type (StorageTypes): the new storage type to convert to

        # Returns
        Dataframe: this Dataframe or a new Dataframe
        """
        col_name = _parse_column_names(self, column)[0]
        return self._df_command('withstoragetypes', df=self.id,
                                storageTypes={col_name: storage_type.to_json()})

    def with_variable_type(self, column, variable_type=VariableTypes.DISCRETE):
        """
        Manually set the variable type of the given column.
        Raise error if the new variable type is not compatible with the storage type of the column.

        # Arguments
        column: a string or a Column object
        variable_type (VariableTypes): the new variable type

        # Returns
        Dataframe: this Dataframe or a new Dataframe
        """
        col_name = _parse_column_names(self, column)[0]
        return self._df_command('withvariabletypes', df=self.id,
                                variableTypes={col_name: variable_type.to_json()})

    """
    Sampling functions
    """

    def take(self, n=10):
        """
        Take a sample of this ``Dataframe``

        # Arguments
        n (int): maximum number of rows to be taken

        # Returns
        DataSample: sample of maximum size `n`
        """
        r = self._client.post_and_wait('df/take', {'df': self.id, 'n': n})
        return DataSample.from_json(r)

    def sample(self, prob=0.1, replacement=True, seed=42):
        """
        Take a sample from this ``Dataframe`` with or without replacement at the given probability.

        Note that this function is probabilistic: it samples each row with a probability of ``prob``,
        therefore if the original ``Dataframe`` has ``N`` rows, the result of this function
        will *NOT* have exactly ``N * prob`` rows.

        # Arguments
        prob (float): The probability to sample each row in the ``Dataframe``
        replacement (bool): Whether to sample with replacement
        seed (int): random seed

        # Returns
        Dataframe: a sample
        """
        return self._df_command('sample', df=self.id, fraction=prob,
                                withReplacement=replacement, seed=seed)

    def show(self, n=5):
        """
        Convenient function to show basic information and sample rows from this Dataframe

        # Arguments
        n (int): Number of rows to show

        # Example

        ```python
        df.groupby(df.customer, 'cylinder_number').count().show()
            ID: 4ae73a7c-27e5-43b9-b69f-43c606e38ee0
            Shape: (490, 3)
            Sample 5 rows:
                customer cylinder_number  count
            0     MODMAT            F108      1
            1     MODMAT              M4      1
            2  CASLIVING             M74      1
            3    ECKERDS            X381      1
            4     TARGET             R43      1
        ```
        """
        pandas_df = self.take(n).to_pandas()
        print('ID: {}\nShape: {}\nSample {} rows:\n{!r}'.format(
            self.id, self.shape, len(pandas_df), pandas_df))

    """
    SQL API
    """

    def select(self, *columns):
        """
        Selects a set of columns based on expressions.

        # Arguments
        columns: list of columns, or column names

        # Returns
        Dataframe: result of the select operation

        # Example

        ```python

        import pycebes as cb

        # different ways to provide arguments to `select`
        df.select(df.customer, cb.substring('cylinder_number', 0, 1).alias('cylinder_t'),
                  cb.col('hardener'), 'wax').show()
            ...
              customer cylinder_t  hardener  wax
            0  TVGUIDE          X       1.0  2.5
            1  TVGUIDE          X       0.7  2.5
            2   MODMAT          B       0.9  2.8
            3   MASSEY          T       1.3  2.5
            4    KMART          J       0.6  2.3

        # Select all columns in a dataframe can be done in several ways
        df.select('*')
        df.select(df['*'])
        df.select(cb.col('*'))
        df.alias('my_name').select('my_name.*')
        ```
        """
        columns = _parse_columns(self, *columns)
        return self._df_command('select', df=self.id, cols=[col.to_json() for col in columns])

    def where(self, condition):
        """
        Filters rows using the given condition.

        # Arguments
        condition (Column): the condition as a Column

        # Example
        ```python
        df.where((df.hardener >= 1) & (df.wax < 2.8)).select('wax', df.hardener, df.customer).show()
            ...
               wax  hardener     customer
            0  2.5       1.0      TVGUIDE
            1  2.5       1.3       MASSEY
            2  2.5       1.1        ROSES
            3  2.5       1.0  HANOVRHOUSE
            4  2.0       1.0   GUIDEPOSTS
        ```
        """
        require(isinstance(condition, Column), 'condition: expect a Column object')
        return self._df_command('where', df=self.id, cols=[condition.to_json()])

    def limit(self, n=100):
        """
        Returns a new ``Dataframe`` by taking the first ``n`` rows.
        """
        return self._df_command('limit', df=self.id, n=n)

    def intersect(self, other):
        """
        Returns a new Dataframe containing rows only in both this Dataframe and another Dataframe.

        # Arguments
        other (Dataframe): another Dataframe to compute the intersection

        # Example
        ```python
        df.where(df.wax > 2).intersect(df.where(df.wax < 2.8)).select('customer', 'wax').show()
            ...
                    customer  wax
            0      SERVMERCH  2.7
            1        TVGUIDE  2.5
            2         MODMAT  2.7
            3           AMES  2.4
            4  colorfulimage  2.5
        ```
        """
        return self._df_command('intersect', df=self.id, otherDf=other.id)

    def union(self, other):
        """
        Returns a new Dataframe containing union of rows in this Dataframe and another Dataframe
        (without deduplication)

        # Arguments
        other (Dataframe): another Dataframe to compute the union

        #Example:
        ```python
        df.where(df.wax < 2).union(df.where(df.wax > 2.8)).select('customer', 'wax').show()
            ...
                customer  wax
            0      USCAV  1.1
            1   COLORTIL  1.7
            2   COLORTIL  1.0
            3  ABBYPRESS  1.0
            4  ABBYPRESS  1.0
        ```
        """
        return self._df_command('union', df=self.id, otherDf=other.id)

    def subtract(self, other):
        """
        Returns a new Dataframe containing rows in this Dataframe but not in another Dataframe.
        This is equivalent to `EXCEPT` in SQL.

        # Arguments
        other (Dataframe): another Dataframe to compute the except

        # Example
        ```python
        df.where(df.wax < 2.8).subtract(df.where(df.wax < 2)).select('customer', 'wax').show()
            ...
                    customer  wax
            0      SERVMERCH  2.7
            1        TVGUIDE  2.5
            2         MODMAT  2.7
            3           AMES  2.4
            4  colorfulimage  2.5
        ```
        """
        return self._df_command('except', df=self.id, otherDf=other.id)

    def join(self, other, expr, join_type='inner'):
        """
        Join with another ``Dataframe``, using the given join expression.

        # Arguments
        other (Dataframe): right side of the join
        expr (Column): Join expression, as a ``Column``
        join_type (str): One of: `inner`, `outer`, `left_outer`, `right_outer`, `leftsemi`, `leftanti`, `cross`

        # Example
        ```python
        df1.join(df2, df1.df1Key == df2.df2Key, join_type='outer')

        # for self-joins, you should rename the columns so that they are accessible after the join
        df1 = df.where(df.wax > 2).select(df[c].alias('df1_{}'.format(c)) for c in df.columns)
        df2 = df.where(df.wax < 2.2).select(df[c].alias('df2_{}'.format(c)) for c in df.columns)

        df_join = df1.join(df2, df1.df1_customer == df2.df2_customer)
        ```
        """
        join_types = ('inner', 'outer', 'left_outer', 'right_outer', 'leftsemi', 'leftanti', 'cross')
        join_type = join_type.lower()
        require(join_type in join_types,
                'Invalid join type: {}. Valid values are: {}'.format(join_type, ', '.join(join_types)))

        return self._df_command('join', leftDf=self.id, rightDf=other.id,
                                joinExprs=expr.to_json(), joinType=join_type)

    @property
    def broadcast(self):
        """
        Marks a Dataframe as small enough for use in broadcast joins.
        """
        return self._df_command('broadcast', df=self.id)

    def alias(self, alias='new_name'):
        """
        Returns a new Dataframe with an alias set
        """
        return self._df_command('alias', df=self.id, alias=alias)

    def with_column(self, col_name, col):
        """
        Returns a new ``Dataframe`` by adding a column or replacing
        the existing column that has the same name (case-insensitive).

        # Arguments
        col_name (str): new column name
        col (Column): ``Column`` object describing the new column
        """
        return self._df_command('withcolumn', df=self.id, colName=col_name, col=col.to_json())

    def with_column_renamed(self, existing_name, new_name):
        """
        Returns a new ``Dataframe`` with a column renamed.

        # Arguments
        existing_name (str):
        new_name (str):
        """
        return self._df_command('withcolumnrenamed', df=self.id, existingName=existing_name, newName=new_name)

    def groupby(self, *columns):
        """
        Groups the ``Dataframe`` using the specified columns, so that we can run aggregation on them.

        See #GroupedDataframe for all the available aggregate functions.

        # Arguments
        columns: list of column names or ``Column`` objects

        # Returns
        GroupedDataframe: object providing aggregation functions

        # Example
        ```python
        df.groupby(df.customer, 'cylinder_number').count().show()
            ...
                customer cylinder_number  count
            0     MODMAT            F108      1
            1     MODMAT              M4      1
            2  CASLIVING             M74      1
            3    ECKERDS            X381      1
            4     TARGET             R43      1

        df.groupby().count().show()
            ...
               count
            0    540
        ```

        """
        return GroupedDataframe(df=self, agg_columns=_parse_columns(self, *columns),
                                agg_type=GroupedDataframe.GROUPBY)

    def rollup(self, *columns):
        """
        Create a multi-dimensional rollup for the current ``Dataframe`` using the specified columns,
        so we can run aggregation on them.

        See #GroupedDataframe for all the available aggregate functions.

        # Arguments
        columns: list of column names or ``Column`` objects

        # Returns
        GroupedDataframe: object providing aggregation functions

        # Example
        ```python
        df.rollup(df.customer, 'proof_on_ctd_ink').count().show()
            ...
                  customer proof_on_ctd_ink  count
            0  NTLWILDLIFE              YES      1
            1     HANHOUSE              YES      2
            2   YIELDHOUSE              YES      1
            3      toysrus                       3
            4          CVS                       2

        df.rollup(df.customer, 'proof_on_ctd_ink').agg({'hardener': 'max', 'wax': 'avg'}).show()
            ...
                  customer proof_on_ctd_ink  max(hardener)  avg(wax)
            0  NTLWILDLIFE              YES            0.6      3.00
            1     HANHOUSE              YES            1.0      2.25
            2   YIELDHOUSE              YES            0.5      3.00
            3      toysrus                             2.1      2.40
            4          CVS                             1.0      2.30
        ```
        """
        return GroupedDataframe(df=self, agg_columns=_parse_columns(self, *columns),
                                agg_type=GroupedDataframe.ROLLUP)

    def cube(self, *columns):
        """
        Create a multi-dimensional cube for the current ``Dataframe`` using the specified columns,
        so we can run aggregation on them.

        See #GroupedDataframe for all the available aggregate functions.

        # Arguments
        columns: list of column names or ``Column`` objects

        # Returns
        GroupedDataframe: object providing aggregation functions

        # Example
        ```python
        df.cube(df.customer, 'proof_on_ctd_ink').count().show()
            ...
                  customer proof_on_ctd_ink  count
            0  NTLWILDLIFE              YES      1
            1     HANHOUSE              YES      2
            2   YIELDHOUSE              YES      1
            3      toysrus                       3
            4          CVS                       2

        df.cube(df.customer, 'proof_on_ctd_ink').agg({'hardener': 'max', 'wax': 'avg'}).show()
            ...
                  customer proof_on_ctd_ink  max(hardener)  avg(wax)
            0  NTLWILDLIFE              YES            0.6      3.00
            1     HANHOUSE              YES            1.0      2.25
            2   YIELDHOUSE              YES            0.5      3.00
            3      toysrus                             2.1      2.40
            4          CVS                             1.0      2.30
        ```
        """
        return GroupedDataframe(df=self, agg_columns=_parse_columns(self, *columns),
                                agg_type=GroupedDataframe.CUBE)

    def agg(self, *exprs):
        """
        Compute aggregates and returns the result as a DataFrame.

        This is a convenient method in which the aggregations are computed on
        all rows. In other words, ``self.agg(*exprs)`` is equivalent to ``self.groupby().agg(*exprs)``

        If exprs is a single dict mapping from string to string,
        then the key is the column to perform aggregation on, and the value is the aggregate function.

        The available aggregate functions are `avg`, `max`, `min`, `sum`, `count`.

        Alternatively, exprs can also be a list of aggregate ``Column`` expressions.

        # Arguments
        exprs (dict, list): a dict mapping from column name (string) to aggregate functions (string),
            or a list of #Column

        # Example
        ```python
        import pycebes as cb

        # count number of non-NA values in column `hardener`
        df.agg(cb.count(df.hardener)).show()
                count(hardener)
            0              533

        # count number of non-NA values in all columns:
        df.agg(*[cb.count(c).alias(c) for c in df.columns])
        ```

        See #GroupedDataframe for more examples
        """
        return self.groupby().agg(*exprs)

    """
    Exploratory functions
    """

    def sort(self, *args):
        """
        Sort this ``Dataframe`` based on the given list of expressions.

        # Arguments
        args: a list of arguments to specify how to sort the Dataframe, where each element
            is either a ``Column`` object or a column name.
            When the element is a column name, it will be sorted in ascending order.

        # Returns
         Dataframe: a new, sorted `Dataframe`

        # Example
        ```python
        df2 = df1.sort()
        df2 = df1.sort(df1['timestamp'].asc, df1['customer'].desc)
        df2 = df1.sort('timestamp', df1.customer.desc)

        # raise ValueError
        df1.sort('non_exist')
        ```
        """
        cols = []
        col_names = set(self.columns)
        for c in args:
            if isinstance(c, six.text_type):
                require(c in col_names, 'Column not found: {}'.format(c))
                cols.append(self[c].asc.to_json())
            else:
                require(isinstance(c, Column), 'Expect a column or a column name, got {!r}'.format(c))
                cols.append(c.to_json())

        return self._df_command('sort', df=self.id, cols=cols)

    def drop(self, *columns):
        """
        Drop columns in this ``Dataframe``

        # Arguments
        columns: column names or #Column objects

        # Returns
        Dataframe: a new ``Dataframe`` with the given columns dropped.
            If ``columns`` is empty, this ``Dataframe`` is returned

        # Example
        ```python
        df1.show()
            ...
              customer cylinder_number  hardener
            0  TVGUIDE            X126       1.0
            1  TVGUIDE            X266       0.7
            2   MODMAT              B7       0.9
            3   MASSEY            T133       1.3
            4    KMART             J34       0.6

        df1.drop(df1.customer, 'hardener').show()
            ...
              cylinder_number
            0            X126
            1            X266
            2              B7
            3            T133
            4             J34
        ```
        """
        col_names = _parse_column_names(self, *columns)
        if len(col_names) == 0:
            return self
        return self._df_command('dropcolumns', df=self.id, colNames=col_names)

    def drop_duplicates(self, *columns):
        """
        Returns a new Dataframe that contains only the unique rows from this Dataframe, only considered
        the given list of columns.

        If no columns are given (the default), all columns will be considered

        # Returns
        Dataframe: a new ``Dataframe`` with duplicated rows removed

        # Example
        ```python
        # original Dataframe has 540 rows
        df.shape
            (540, 40)

        # No duplicated rows in this Dataframe, therefore the shape stays the same
        df.drop_duplicates().shape
            (540, 40)

        # only consider the `customer` column
        df.drop_duplicates(df.customer).shape
            (83, 40)

        # we can check by computing the number of distinct values in column `customer`
        df.select(cb.count_distinct(df.customer)).take().to_pandas()
               count(DISTINCT customer)
            0                        83

        # only consider 2 columns
        df.drop_duplicates(df.customer, 'cylinder_number').shape
            (490, 40)
        ```
        """
        col_names = _parse_column_names(self, *columns)
        if len(col_names) == 0:
            col_names = self.columns
        return self._df_command('dropduplicates', df=self.id, colNames=col_names)

    def dropna(self, how='any', thresh=None, columns=None):
        """
        Return a new Dataframe with rows omitted where some of the data are missing

        # Arguments
        how (str): strategy to drop rows. Accepted values are:
            * ``any``: if any NA values are present, drop that label
            * ``all``: if all values are NA, drop that label
        thresh (int): drop rows containing less than ``thresh`` non-null and non-NaN values
            If ``thresh`` is specified, ``how`` will be ignored.
        columns (list): a list of columns to include. Default is None, meaning all columns are included

        # Returns
        Dataframe: a new ``Dataframe`` with NA values dropped

        # Example
        ```python
        df.dropna()
        df.dropna(how='all')
        df.dropna(thresh=38)
        df.dropna(thresh=2, columns=[df.wax, 'proof_on_ctd_ink', df.hardener])
        ```
        """
        if columns is None:
            columns = self.columns
        columns = _parse_column_names(self, *columns)

        min_non_null = thresh
        if min_non_null is None:
            require(how in ('any', 'all'), 'how={} is not supported. Only "any" or "all" are allowed'.format(how))
            if how == 'any':
                min_non_null = len(columns)
            else:
                min_non_null = 1
        return self._df_command('dropna', df=self.id, minNonNulls=min_non_null, colNames=columns)

    def fillna(self, value=None, columns=None):
        """
        Fill NA/NaN values using the specified value.

        # Arguments
        value: the value used to fill the holes. Can be a ``float``, ``string`` or a ``dict``.
            When ``value`` is a dict, it is a map from column names to a value used to fill the holes in that column.
            In this case the value can only be of type ``int``, ``float``, ``string`` or ``bool``.
            The values will be casted to the column data type.
        columns (list): List of columns to consider. If None, all columns are considered

        # Returns:
        Dataframe: a new Dataframe with holes filled
        """
        if columns is None:
            columns = self.columns
        columns = _parse_column_names(self, *columns)
        if isinstance(value, int):
            value = float(value)
        if isinstance(value, (six.text_type, float)):
            return self._df_command('fillna', df=self.id, value=value, colNames=columns)
        if isinstance(value, dict):
            return self._df_command('fillnawithmap', df=self.id, valueMap=to_json(value))
        raise ValueError('Unsupported value: {!r}'.format(value))


@six.python_2_unicode_compatible
class GroupedDataframe(object):
    """
    Represent a grouped Dataframe, providing aggregation functions.
    """

    GROUPBY = 'GroupBy'
    ROLLUP = 'RollUp'
    CUBE = 'Cube'

    def __init__(self, df, agg_columns=(), agg_type=GROUPBY, pivot_column=None, pivot_values=()):
        """
        Construct a new GroupDataframe object. Not to be used by end-users.

        # Arguments
        df (Dataframe):
        agg_columns: list of aggregate columns
        agg_type: aggregation type
        pivot_column: column name of pivot
        pivot_values: pivot values
        """
        self.df = df
        self.agg_type = agg_type
        self.agg_columns = agg_columns
        self.pivot_column = pivot_column
        self.pivot_values = pivot_values

    def __repr__(self):
        return ('{}(df_id={!r}, agg_columns={!r}, agg_type={!r}, pivot_column={!r}, '
                'pivot_values={!r})'.format(self.__class__.__name__,
                                            self.df.id, self.agg_columns,
                                            self.agg_type, self.pivot_column, self.pivot_values))

    def _send_request(self, generic_agg_exprs=None, agg_func=None, agg_col_names=None):
        client = get_default_session().client

        data = {
            'df': self.df.id,
            'cols': [c.to_json() for c in self.agg_columns],
            'aggType': self.agg_type,
            'pivotValues': None,
            'aggColNames': []
        }
        if self.pivot_column is not None:
            data['pivotColName'] = self.pivot_column
        if self.pivot_values is not None:
            data['pivotValues'] = [to_json(v) for v in self.pivot_values]
        if generic_agg_exprs is not None:
            data['genericAggExprs'] = [c.to_json() for c in generic_agg_exprs]
        if agg_func is not None:
            data['aggFunc'] = agg_func
        if agg_col_names is not None:
            data['aggColNames'] = _parse_column_names(self.df, *agg_col_names)

        return Dataframe.from_json(client.post_and_wait('df/aggregate', data))

    def agg(self, *exprs):
        """
        Compute aggregates and returns the result as a DataFrame.

        If exprs is a single dict mapping from string to string,
        then the key is the column to perform aggregation on, and the value is the aggregate function.

        The available aggregate functions are `avg`, `max`, `min`, `sum`, `count`.

        Alternatively, exprs can also be a list of aggregate ``Column`` expressions.

        # Arguments
        exprs (dict, list): a dict mapping from column name (string) to aggregate functions (string),
            or a list of :class:`Column`

        # Example
        ```python
        import pycebes as cb

        df.groupby(df.customer).agg(cb.max(df.timestamp), cb.min('job_number'),
                                    cb.count_distinct('job_number').alias('cnt')).show()
            ...
                   customer  max(timestamp)  min(job_number)  cnt
                  WOOLWORTH        19920405            34227    5
                   HOMESHOP        19910129            38064    1
               homeshopping        19920712            36568    1
                     GLOBAL        19900621            36846    1
                        JCP        19910322            34781    2

        df.groupby(df.customer).agg({'timestamp': 'max', 'job_number': 'min', 'job_number': 'count'}).show()
            ...
                   customer  max(timestamp)  count(job_number)
                  WOOLWORTH        19920405                  9
                   HOMESHOP        19910129                  2
               homeshopping        19920712                  1
                     GLOBAL        19900621                  1
                        JCP        19910322                  4
        ```
        """
        agg_cols = []

        if len(exprs) == 1 and isinstance(exprs[0], dict):
            funcs = {'avg': functions.avg, 'max': functions.max,
                     'min': functions.min, 'sum': functions.sum,
                     'count': functions.count}

            for col, func in exprs[0].items():
                require(func in funcs, 'Unsupported aggregation function: {!r}'.format(func))
                agg_cols.append(funcs[func](functions.col(col)))
        else:
            for expr in exprs:
                require(isinstance(expr, Column), 'Expected a Column expression, got {!r}'.format(expr))
                agg_cols.append(expr)

        return self._send_request(generic_agg_exprs=agg_cols)

    def count(self):
        """
        Counts the number of records for each group.

        # Example
        ```python
        df.groupby(df.customer).count().show()
            ...
                  customer  count
                 WOOLWORTH      9
                  HOMESHOP      2
              homeshopping      1
                    GLOBAL      1
                       JCP      4

        df.groupby(df.customer, 'proof_on_ctd_ink').count().show()
            ...
                   customer proof_on_ctd_ink  count
                 GUIDEPOSTS              YES      5
                   homeshop                       4
              colorfulimage                       1
                  CASLIVING              YES      2
                      ABBEY              YES      4
        ```
        """
        return self._send_request(agg_func='count')

    def min(self, *columns):
        """
        Computes the min value for each numeric column for each group.

        # Arguments
        columns: list of column names (string). Non-numeric columns are ignored.

        # Example
        ```python
        df.groupby(df.customer).min('hardener', 'wax').show()
            ...
                  customer  min(hardener)  min(wax)
                 WOOLWORTH            0.0      0.00
                  HOMESHOP            0.8      2.50
              homeshopping            NaN      2.50
                    GLOBAL            1.1      3.00
                       JCP            0.6      1.75
        ```
        """
        return self._send_request(agg_func='min', agg_col_names=columns)

    def max(self, *columns):
        """
        Computes the max value for each numeric column for each group.

        # Arguments
        columns: list of column names (string). Non-numeric columns are ignored.

        # Example
        ```python
        df.groupby(df.customer).max('hardener', 'wax').show()
            ...
                   customer  max(hardener)  max(wax)
                  WOOLWORTH            1.3       2.6
                   HOMESHOP            1.8       2.5
               homeshopping            NaN       2.5
                     GLOBAL            1.1       3.0
                        JCP            1.7       3.0
        ```
        """
        return self._send_request(agg_func='max', agg_col_names=columns)

    def mean(self, *columns):
        """
        Computes the average value for each numeric column for each group.

        # Arguments
        columns: list of column names (string). Non-numeric columns are ignored.

        # Example
        ```python
        df.groupby(df.customer).mean('hardener', 'wax').show()
            ...
                   customer  avg(hardener)  avg(wax)
                  WOOLWORTH       0.811111  1.744444
                   HOMESHOP       1.300000  2.500000
               homeshopping            NaN  2.500000
                     GLOBAL       1.100000  3.000000
                        JCP       0.975000  2.437500
        ```
        """
        return self._send_request(agg_func='mean', agg_col_names=columns)

    avg = mean

    def sum(self, *columns):
        """
        Computes the sum for each numeric column for each group.

        # Arguments
        columns: list of column names (string). Non-numeric columns are ignored.

        # Example
        ```python
        df.groupby(df.customer).sum('hardener', 'wax').show()
            ...
                   customer  sum(hardener)  sum(wax)
                  WOOLWORTH            7.3     15.70
                   HOMESHOP            2.6      5.00
               homeshopping            NaN      2.50
                     GLOBAL            1.1      3.00
                        JCP            3.9      9.75
        ```
        """
        return self._send_request(agg_func='sum', agg_col_names=columns)

    def pivot(self, column, values=None):
        """
        Pivots a column of the ``Dataframe`` and perform the specified aggregation.

        # Arguments
        column (str): Name of the column to pivot.
        values: List of values that will be translated to columns in the output ``Dataframe``
            If unspecified, Cebes will need to compute the unique values of the given column before
            the pivot, therefore it will be less efficient.

        # Returns
        GroupedDataframe: another `GroupedDataframe` object, on which you can call aggregation functions

        # Example
        ```python
        # for each pair of (`customer`, `proof_on_ctd_ink`), count the number of entries
        df.groupby(df.customer).pivot('proof_on_ctd_ink').count().show()
            ...
                   customer        NO  YES
                  WOOLWORTH  1.0  1.0  7.0
                   HOMESHOP  NaN  NaN  2.0
               homeshopping  1.0  NaN  NaN
                     GLOBAL  NaN  NaN  1.0
                        JCP  NaN  NaN  4.0

        # for each pair of (`customer`, `proof_on_ctd_ink`), get the max value of `hardener`
        df.groupby(df.customer).pivot('proof_on_ctd_ink').max('hardener').show()
            ...
                  customer        NO  YES
                 WOOLWORTH  1.0  0.0  1.3
                  HOMESHOP  NaN  NaN  1.8
              homeshopping  NaN  NaN  NaN
                    GLOBAL  NaN  NaN  1.1
                       JCP  NaN  NaN  1.7

        # specify the pivot values, more efficient
        df.groupby(df.customer).pivot('proof_on_ctd_ink', values=['YES', 'NO']).count().show()
            ...
                   customer  YES   NO
                  WOOLWORTH  7.0  1.0
                   HOMESHOP  2.0  NaN
               homeshopping  NaN  NaN
                     GLOBAL  1.0  NaN
                        JCP  4.0  NaN
        ```
        """
        return GroupedDataframe(df=self.df, agg_columns=self.agg_columns,
                                agg_type=self.agg_type, pivot_column=column,
                                pivot_values=values)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from collections import namedtuple

import six

from pycebes.core.schema import StorageType
from pycebes.internal.serializer import to_json

_ParamConfig = namedtuple('_ParamConfig', ['name', 'param_type', 'server_name'])


def param(name='param', param_type=None, server_name=None):
    """
    Decorator used for parameters of Expressions

    :param name: Name of the parameter (i.e. slot) 
    :param param_type: optionally, a string represent the type of this param (on the server)
        mostly used to disambiguous primitive types on server
    :type param_type: str
    :param server_name: name of this parameter on server, if different from 'name'
    :return: 
    """

    def decorate(cls):
        assert issubclass(cls, Expression)
        if getattr(cls, 'PARAMS', None) is None:
            cls.PARAMS = {}

        class_name = cls.__name__
        if class_name not in cls.PARAMS:
            cls.PARAMS[class_name] = []

        pc = _ParamConfig(name, param_type, name if server_name is None else server_name)
        if next((p for p in cls.PARAMS[class_name] if p.name == pc.name), None) is not None:
            raise ValueError('Duplicated parameter named {} in class {}'.format(pc.name, cls.__name__))
        cls.PARAMS[class_name].append(pc)
        return cls

    return decorate


def server_namespace(ns='io.cebes.df.expressions'):
    """
    Decorator used for Expressions, to specify the namespace of this class on the server
    """

    def decorate(cls):
        assert issubclass(cls, Expression)
        setattr(cls, '_server_namespace', ns)
        return cls

    return decorate


############################################################################


@six.python_2_unicode_compatible
class Expression(object):
    def __init__(self, **kwargs):
        param_names = {pc.name for pc in self._get_params()}
        for k, v in kwargs.items():
            if k not in param_names:
                raise ValueError('Invalid param name {} in class {}'.format(k, self.__class__.__name__))
            setattr(self, k, v)

    def to_json(self):
        js = {'className': '{}.{}'.format(self._get_server_namespace(), self.__class__.__name__)}
        for pc in self._get_params():
            assert isinstance(pc, _ParamConfig)
            value = getattr(self, pc.name, None)
            js[pc.server_name] = self._param_to_json(pc, value)
        return js

    def __repr__(self):
        params_str = ','.join('{}={!r}'.format(pc.name, getattr(self, pc.name, None)) for pc in self._get_params())
        return '{}({})'.format(self.__class__.__name__, params_str)

    @staticmethod
    def _param_to_json(pc, value):
        """
        Serialize a parameter to JSON, can be overriden by subclasses
        :type pc: _ParamConfig
        :return: a JSON value
        """
        return value.to_json() if isinstance(value, Expression) else to_json(value, pc.param_type)

    def _get_params(self):
        """
        Return the list of params of this expression
        :rtype: list[_ParamConfig]
        """
        params = []
        for parent_class in self.__class__.__mro__:
            params.extend(self.PARAMS.get(parent_class.__name__, []))
        return params

    @classmethod
    def _get_server_namespace(cls):
        return getattr(cls, '_server_namespace', 'io.cebes.df.expressions')


############################################################################
# Helpers subclasses of Expression

@param('children')
class _WithChildren(Expression):
    """
    Subclass of Expression with serialization logic for "children" parameter,
    which is a list of Expressions
    """

    def __init__(self, **kwargs):
        super(_WithChildren, self).__init__(**kwargs)

    @staticmethod
    def _param_to_json(pc, value):
        if pc.name == 'children':
            assert isinstance(value, (list, tuple)) and all(isinstance(ex, Expression) for ex in value)
            return [ex.to_json() for ex in value]
        raise ValueError('Missing serialization logic for {} of class {!r}'.format(pc.name, value))


@param('child')
class _UnaryExpression(Expression):
    """
    Subclass of Expression with serialization logic for "child" parameter
    """

    def __init__(self, child, **kwargs):
        super(_UnaryExpression, self).__init__(child=child, **kwargs)


@param('left')
@param('right')
class _BinaryExpression(Expression):
    """
    Subclass of Expression with serialization logic for "left" and "right" parameters
    """

    def __init__(self, left, right):
        super(_BinaryExpression, self).__init__(left=left, right=right)


"""
Aggregation
"""


@param('relative_sd', param_type='double', server_name='relativeSD')
class ApproxCountDistinct(_UnaryExpression):
    def __init__(self, child, relative_sd=0.05):
        super(ApproxCountDistinct, self).__init__(child=child, relative_sd=relative_sd)


class Average(_UnaryExpression):
    def __init__(self, child):
        super(Average, self).__init__(child=child)


class CollectList(_UnaryExpression):
    def __init__(self, child):
        super(CollectList, self).__init__(child=child)


class CollectSet(_UnaryExpression):
    def __init__(self, child):
        super(CollectSet, self).__init__(child=child)


class Corr(_BinaryExpression):
    def __init__(self, left, right):
        super(Corr, self).__init__(left=left, right=right)


class Count(_UnaryExpression):
    def __init__(self, child):
        super(Count, self).__init__(child=child)


@param('expr')
@param('exprs')
class CountDistinct(Expression):
    def __init__(self, *exprs):
        if len(exprs) == 0:
            raise ValueError('Empty list of expressions for {}'.format(self.__class__.__name__))
        expr = exprs[0]
        exprs = exprs[1:]
        super(CountDistinct, self).__init__(expr=expr, exprs=exprs)

    @staticmethod
    def _param_to_json(pc, value):
        if pc.name == 'expr':
            assert isinstance(value, Expression)
            return value.to_json()
        if pc.name == 'exprs':
            return [v.to_json() for v in value]
        return Expression._param_to_json(pc, value)


class CovPopulation(_BinaryExpression):
    def __init__(self, left, right):
        super(CovPopulation, self).__init__(left=left, right=right)


class CovSample(_BinaryExpression):
    def __init__(self, left, right):
        super(CovSample, self).__init__(left=left, right=right)


@param('ignore_nulls', param_type=bool, server_name='ignoreNulls')
class First(_UnaryExpression):
    def __init__(self, child, ignore_nulls=False):
        super(First, self).__init__(child=child, ignore_nulls=ignore_nulls)


class Grouping(_UnaryExpression):
    def __init__(self, child):
        super(Grouping, self).__init__(child=child)


class GroupingID(_WithChildren):
    def __init__(self, children=()):
        super(GroupingID, self).__init__(children=children)


class Kurtosis(_UnaryExpression):
    def __init__(self, child):
        super(Kurtosis, self).__init__(child=child)


@param('ignore_nulls', param_type=bool, server_name='ignoreNulls')
class Last(_UnaryExpression):
    def __init__(self, child, ignore_nulls=False):
        super(Last, self).__init__(child=child, ignore_nulls=ignore_nulls)


class Max(_UnaryExpression):
    def __init__(self, child):
        super(Max, self).__init__(child=child)


class Min(_UnaryExpression):
    def __init__(self, child):
        super(Min, self).__init__(child=child)


class Skewness(_UnaryExpression):
    def __init__(self, child):
        super(Skewness, self).__init__(child=child)


class StddevSamp(_UnaryExpression):
    def __init__(self, child):
        super(StddevSamp, self).__init__(child=child)


class StddevPop(_UnaryExpression):
    def __init__(self, child):
        super(StddevPop, self).__init__(child=child)


@param('is_distinct', param_type=bool, server_name='isDistinct')
class Sum(_UnaryExpression):
    def __init__(self, child, is_distinct=False):
        super(Sum, self).__init__(child=child, is_distinct=is_distinct)


class VarianceSamp(_UnaryExpression):
    def __init__(self, child):
        super(VarianceSamp, self).__init__(child=child)


class VariancePop(_UnaryExpression):
    def __init__(self, child):
        super(VariancePop, self).__init__(child=child)


"""
alias
"""


@param('alias')
class Alias(_UnaryExpression):
    def __init__(self, child, alias='alias'):
        super(Alias, self).__init__(child=child, alias=alias)


@param('aliases')
class MultiAlias(_UnaryExpression):
    def __init__(self, child, aliases=()):
        super(MultiAlias, self).__init__(child=child, aliases=aliases)


@param('to')
class Cast(_UnaryExpression):
    def __init__(self, child, to):
        super(Cast, self).__init__(child=child, to=to)

    @staticmethod
    def _param_to_json(pc, value):
        if pc.name == 'child':
            assert isinstance(value, Expression)
            return value.to_json()

        assert pc.name == 'to' and isinstance(value, Literal) and isinstance(value.value, StorageType)
        return value.value.to_json()


"""
arithmetic
"""


class UnaryMinus(_UnaryExpression):
    def __init__(self, child):
        super(UnaryMinus, self).__init__(child=child)


class Add(_BinaryExpression):
    def __init__(self, left, right):
        super(Add, self).__init__(left=left, right=right)


class Subtract(_BinaryExpression):
    def __init__(self, left, right):
        super(Subtract, self).__init__(left=left, right=right)


class Multiply(_BinaryExpression):
    def __init__(self, left, right):
        super(Multiply, self).__init__(left=left, right=right)


class Divide(_BinaryExpression):
    def __init__(self, left, right):
        super(Divide, self).__init__(left=left, right=right)


class Remainder(_BinaryExpression):
    def __init__(self, left, right):
        super(Remainder, self).__init__(left=left, right=right)


class Abs(_UnaryExpression):
    def __init__(self, child):
        super(Abs, self).__init__(child=child)


class Sqrt(_UnaryExpression):
    def __init__(self, child):
        super(Sqrt, self).__init__(child=child)


class Acos(_UnaryExpression):
    def __init__(self, child):
        super(Acos, self).__init__(child=child)


class Asin(_UnaryExpression):
    def __init__(self, child):
        super(Asin, self).__init__(child=child)


class Atan(_UnaryExpression):
    def __init__(self, child):
        super(Atan, self).__init__(child=child)


class Atan2(_BinaryExpression):
    def __init__(self, left, right):
        super(Atan2, self).__init__(left=left, right=right)


class Bin(_UnaryExpression):
    def __init__(self, child):
        super(Bin, self).__init__(child=child)


class Cbrt(_UnaryExpression):
    def __init__(self, child):
        super(Cbrt, self).__init__(child=child)


class Ceil(_UnaryExpression):
    def __init__(self, child):
        super(Ceil, self).__init__(child=child)


@param('from_base', param_type='int', server_name='fromBase')
@param('to_base', param_type='int', server_name='toBase')
class Conv(_UnaryExpression):
    def __init__(self, child, from_base=10, to_base=2):
        super(Conv, self).__init__(child=child, from_base=from_base, to_base=to_base)


class Cos(_UnaryExpression):
    def __init__(self, child):
        super(Cos, self).__init__(child=child)


class Cosh(_UnaryExpression):
    def __init__(self, child):
        super(Cosh, self).__init__(child=child)


class Exp(_UnaryExpression):
    def __init__(self, child):
        super(Exp, self).__init__(child=child)


class Expm1(_UnaryExpression):
    def __init__(self, child):
        super(Expm1, self).__init__(child=child)


class Factorial(_UnaryExpression):
    def __init__(self, child):
        super(Factorial, self).__init__(child=child)


class Floor(_UnaryExpression):
    def __init__(self, child):
        super(Floor, self).__init__(child=child)


class Greatest(_WithChildren):
    def __init__(self, children=()):
        super(Greatest, self).__init__(children=children)


class Hex(_UnaryExpression):
    def __init__(self, child):
        super(Hex, self).__init__(child=child)


class Unhex(_UnaryExpression):
    def __init__(self, child):
        super(Unhex, self).__init__(child=child)


class Hypot(_BinaryExpression):
    def __init__(self, left, right):
        super(Hypot, self).__init__(left=left, right=right)


class Least(_WithChildren):
    def __init__(self, children=()):
        super(Least, self).__init__(children=children)


class Log(_UnaryExpression):
    def __init__(self, child):
        super(Log, self).__init__(child=child)


@param('base', param_type='double')
class Logarithm(_UnaryExpression):
    def __init__(self, child, base=10.0):
        super(Logarithm, self).__init__(child=child, base=base)


class Log10(_UnaryExpression):
    def __init__(self, child):
        super(Log10, self).__init__(child=child)


class Log1p(_UnaryExpression):
    def __init__(self, child):
        super(Log1p, self).__init__(child=child)


class Log2(_UnaryExpression):
    def __init__(self, child):
        super(Log2, self).__init__(child=child)


class Pow(_BinaryExpression):
    def __init__(self, left, right):
        super(Pow, self).__init__(left=left, right=right)


class Pmod(_BinaryExpression):
    def __init__(self, left, right):
        super(Pmod, self).__init__(left=left, right=right)


class Rint(_UnaryExpression):
    def __init__(self, child):
        super(Rint, self).__init__(child=child)


class Round(_BinaryExpression):
    def __init__(self, left, right):
        super(Round, self).__init__(left=left, right=right)


@param('scale', param_type='int')
class BRound(_UnaryExpression):
    def __init__(self, child, scale=5):
        super(BRound, self).__init__(child=child, scale=scale)


@param('num_bits', param_type='int', server_name='numBits')
class ShiftLeft(_UnaryExpression):
    def __init__(self, child, num_bits=5):
        super(ShiftLeft, self).__init__(child=child, num_bits=num_bits)


@param('num_bits', param_type='int', server_name='numBits')
class ShiftRight(_UnaryExpression):
    def __init__(self, child, num_bits=5):
        super(ShiftRight, self).__init__(child=child, num_bits=num_bits)


@param('num_bits', param_type='int', server_name='numBits')
class ShiftRightUnsigned(_UnaryExpression):
    def __init__(self, child, num_bits=5):
        super(ShiftRightUnsigned, self).__init__(child=child, num_bits=num_bits)


class Signum(_UnaryExpression):
    def __init__(self, child):
        super(Signum, self).__init__(child=child)


class Sin(_UnaryExpression):
    def __init__(self, child):
        super(Sin, self).__init__(child=child)


class Sinh(_UnaryExpression):
    def __init__(self, child):
        super(Sinh, self).__init__(child=child)


class Tan(_UnaryExpression):
    def __init__(self, child):
        super(Tan, self).__init__(child=child)


class Tanh(_UnaryExpression):
    def __init__(self, child):
        super(Tanh, self).__init__(child=child)


class ToDegrees(_UnaryExpression):
    def __init__(self, child):
        super(ToDegrees, self).__init__(child=child)


class ToRadians(_UnaryExpression):
    def __init__(self, child):
        super(ToRadians, self).__init__(child=child)


"""
collections
"""


class ArrayContains(_BinaryExpression):
    def __init__(self, left, right):
        super(ArrayContains, self).__init__(left=left, right=right)


class Explode(_UnaryExpression):
    def __init__(self, child):
        super(Explode, self).__init__(child=child)


class PosExplode(_UnaryExpression):
    def __init__(self, child):
        super(PosExplode, self).__init__(child=child)


@param('path')
class GetJsonObject(_UnaryExpression):
    def __init__(self, child, path=''):
        super(GetJsonObject, self).__init__(child=child, path=path)


@param('fields')
class JsonTuple(_UnaryExpression):
    def __init__(self, child, fields=()):
        super(JsonTuple, self).__init__(child=child, fields=fields)


class Size(_UnaryExpression):
    def __init__(self, child):
        super(Size, self).__init__(child=child)


@param('asc')
class SortArray(_UnaryExpression):
    def __init__(self, child, asc=True):
        super(SortArray, self).__init__(child=child, asc=asc)


"""
complex types
"""


class CreateArray(_WithChildren):
    def __init__(self, children=()):
        super(CreateArray, self).__init__(children=children)


class CreateMap(_WithChildren):
    def __init__(self, children=()):
        super(CreateMap, self).__init__(children=children)


class CreateStruct(_WithChildren):
    def __init__(self, children=()):
        super(CreateStruct, self).__init__(children=children)


"""
conditional
"""


@param('branches')
@param('else_value', server_name='elseValue')
class CaseWhen(Expression):
    def __init__(self, branches, else_value=None):
        """
        :param branches: a list [(expr, expr)] where each tuple is a condition and the corresponding value
        :param else_value: another expression for the "else" value, or None if unspecified
        """
        super(CaseWhen, self).__init__(branches=branches, else_value=else_value)

    @staticmethod
    def _param_to_json(pc, value):
        if pc.name == 'branches':
            result = []
            for pair in value:
                result.append([ex.to_json() for ex in pair])
            return result

        assert pc.name == 'else_value'
        assert value is None or isinstance(value, Expression)
        return value.to_json() if isinstance(value, Expression) else None


class GetItem(_BinaryExpression):
    def __init__(self, left, right):
        super(GetItem, self).__init__(left=left, right=right)


@param('field_name', server_name='fieldName')
class GetField(_UnaryExpression):
    def __init__(self, child, field_name=''):
        super(GetField, self).__init__(child=child, field_name=field_name)


"""
datetime
"""


@param('num_months', param_type='int', server_name='numMonths')
class AddMonths(_UnaryExpression):
    def __init__(self, child, num_months=1):
        super(AddMonths, self).__init__(child=child, num_months=num_months)


class CurrentDate(Expression):
    def __init__(self):
        super(CurrentDate, self).__init__()


class CurrentTimestamp(Expression):
    def __init__(self):
        super(CurrentTimestamp, self).__init__()


@param('fmt', server_name='format')
class DateFormatClass(_UnaryExpression):
    def __init__(self, child, fmt=''):
        super(DateFormatClass, self).__init__(child=child, fmt=fmt)


@param('days', param_type='int')
class DateAdd(_UnaryExpression):
    def __init__(self, child, days=1):
        super(DateAdd, self).__init__(child=child, days=days)


@param('days', param_type='int')
class DateSub(_UnaryExpression):
    def __init__(self, child, days=1):
        super(DateSub, self).__init__(child=child, days=days)


class DateDiff(_BinaryExpression):
    def __init__(self, left, right):
        super(DateDiff, self).__init__(left=left, right=right)


class Year(_UnaryExpression):
    def __init__(self, child):
        super(Year, self).__init__(child=child)


class Quarter(_UnaryExpression):
    def __init__(self, child):
        super(Quarter, self).__init__(child=child)


class Month(_UnaryExpression):
    def __init__(self, child):
        super(Month, self).__init__(child=child)


class DayOfMonth(_UnaryExpression):
    def __init__(self, child):
        super(DayOfMonth, self).__init__(child=child)


class DayOfYear(_UnaryExpression):
    def __init__(self, child):
        super(DayOfYear, self).__init__(child=child)


class Hour(_UnaryExpression):
    def __init__(self, child):
        super(Hour, self).__init__(child=child)


class LastDay(_UnaryExpression):
    def __init__(self, child):
        super(LastDay, self).__init__(child=child)


class Minute(_UnaryExpression):
    def __init__(self, child):
        super(Minute, self).__init__(child=child)


class MonthsBetween(_BinaryExpression):
    def __init__(self, left, right):
        super(MonthsBetween, self).__init__(left=left, right=right)


@param('day_of_week', server_name='dayOfWeek')
class NextDay(_UnaryExpression):
    def __init__(self, child, day_of_week=''):
        super(NextDay, self).__init__(child=child, day_of_week=day_of_week)


class Second(_UnaryExpression):
    def __init__(self, child):
        super(Second, self).__init__(child=child)


class WeekOfYear(_UnaryExpression):
    def __init__(self, child):
        super(WeekOfYear, self).__init__(child=child)


@param('fmt', server_name='format')
class FromUnixTime(_UnaryExpression):
    def __init__(self, child, fmt=''):
        super(FromUnixTime, self).__init__(child=child, fmt=fmt)


@param('fmt', server_name='format')
class UnixTimestamp(_UnaryExpression):
    def __init__(self, child, fmt=''):
        super(UnixTimestamp, self).__init__(child=child, fmt=fmt)


class ToDate(_UnaryExpression):
    def __init__(self, child):
        super(ToDate, self).__init__(child=child)


@param('fmt', server_name='format')
class TruncDate(_UnaryExpression):
    def __init__(self, child, fmt=''):
        super(TruncDate, self).__init__(child=child, fmt=fmt)


@param('tz')
class FromUTCTimestamp(_UnaryExpression):
    def __init__(self, child, tz=''):
        super(FromUTCTimestamp, self).__init__(child=child, tz=tz)


@param('tz')
class ToUTCTimestamp(_UnaryExpression):
    def __init__(self, child, tz=''):
        super(ToUTCTimestamp, self).__init__(child=child, tz=tz)


@param('window_duration', server_name='windowDuration')
@param('slide_duration', server_name='slideDuration')
@param('start_time', server_name='startTime')
class TimeWindow(_UnaryExpression):
    def __init__(self, child, window_duration='', slide_duration='', start_time=''):
        super(TimeWindow, self).__init__(child=child, window_duration=window_duration,
                                         slide_duration=slide_duration,
                                         start_time=start_time)


"""
misc
"""


class Md5(_UnaryExpression):
    def __init__(self, child):
        super(Md5, self).__init__(child=child)


class Sha1(_UnaryExpression):
    def __init__(self, child):
        super(Sha1, self).__init__(child=child)


@param('num_bits', param_type='int', server_name='numBits')
class Sha2(_UnaryExpression):
    def __init__(self, child, num_bits=0):
        super(Sha2, self).__init__(child=child, num_bits=num_bits)


class Crc32(_UnaryExpression):
    def __init__(self, child):
        super(Crc32, self).__init__(child=child)


class Murmur3Hash(_WithChildren):
    def __init__(self, children=()):
        super(Murmur3Hash, self).__init__(children=children)


"""
non-aggregation
"""


class Coalesce(_WithChildren):
    def __init__(self, children=()):
        super(Coalesce, self).__init__(children=children)


class InputFileName(Expression):
    def __init__(self):
        super(InputFileName, self).__init__()


class MonotonicallyIncreasingID(Expression):
    def __init__(self):
        super(MonotonicallyIncreasingID, self).__init__()


class NaNvl(_BinaryExpression):
    def __init__(self, left, right):
        super(NaNvl, self).__init__(left=left, right=right)


@param('seed', param_type='long')
class Rand(Expression):
    def __init__(self, seed=42):
        super(Rand, self).__init__(seed=seed)


@param('seed', param_type='long')
class Randn(Expression):
    def __init__(self, seed=42):
        super(Randn, self).__init__(seed=seed)


class SparkPartitionID(Expression):
    def __init__(self):
        super(SparkPartitionID, self).__init__()


@param('expr')
class RawExpression(Expression):
    def __init__(self, expr=''):
        super(RawExpression, self).__init__(expr=expr)


"""
predicates
"""


class Not(_UnaryExpression):
    def __init__(self, child):
        super(Not, self).__init__(child=child)


class EqualTo(_BinaryExpression):
    def __init__(self, left, right):
        super(EqualTo, self).__init__(left=left, right=right)


class GreaterThan(_BinaryExpression):
    def __init__(self, left, right):
        super(GreaterThan, self).__init__(left=left, right=right)


class LessThan(_BinaryExpression):
    def __init__(self, left, right):
        super(LessThan, self).__init__(left=left, right=right)


class GreaterThanOrEqual(_BinaryExpression):
    def __init__(self, left, right):
        super(GreaterThanOrEqual, self).__init__(left=left, right=right)


class LessThanOrEqual(_BinaryExpression):
    def __init__(self, left, right):
        super(LessThanOrEqual, self).__init__(left=left, right=right)


class EqualNullSafe(_BinaryExpression):
    def __init__(self, left, right):
        super(EqualNullSafe, self).__init__(left=left, right=right)


class IsNaN(_UnaryExpression):
    def __init__(self, child):
        super(IsNaN, self).__init__(child=child)


class IsNull(_UnaryExpression):
    def __init__(self, child):
        super(IsNull, self).__init__(child=child)


class IsNotNull(_UnaryExpression):
    def __init__(self, child):
        super(IsNotNull, self).__init__(child=child)


class Or(_BinaryExpression):
    def __init__(self, left, right):
        super(Or, self).__init__(left=left, right=right)


class And(_BinaryExpression):
    def __init__(self, left, right):
        super(And, self).__init__(left=left, right=right)


@param('value')
@param('ls', server_name='list')
class In(Expression):
    def __init__(self, value, ls=()):
        super(In, self).__init__(value=value, ls=ls)

    @staticmethod
    def _param_to_json(pc, value):
        if pc.name == 'value':
            assert isinstance(value, Expression)
            return value.to_json()
        assert pc.name == 'ls'
        assert isinstance(value, (tuple, list)) and all(isinstance(ex, Expression) for ex in value)
        return [ex.to_json() for ex in value]


class BitwiseOr(_BinaryExpression):
    def __init__(self, left, right):
        super(BitwiseOr, self).__init__(left=left, right=right)


class BitwiseAnd(_BinaryExpression):
    def __init__(self, left, right):
        super(BitwiseAnd, self).__init__(left=left, right=right)


class BitwiseXor(_BinaryExpression):
    def __init__(self, left, right):
        super(BitwiseXor, self).__init__(left=left, right=right)


class BitwiseNot(_UnaryExpression):
    def __init__(self, child):
        super(BitwiseNot, self).__init__(child=child)


"""
primitives
"""


@param('value')
class Literal(Expression):
    def __init__(self, value):
        super(Literal, self).__init__(value=value)


@param('col_name', server_name='colName')
class UnresolvedColumnName(Expression):
    def __init__(self, col_name):
        super(UnresolvedColumnName, self).__init__(col_name=col_name)


"""
SortOrder
"""


@param('direction')
class SortOrder(_UnaryExpression):
    Ascending = 'Ascending'
    Descending = 'Descending'

    def __init__(self, child, direction=Ascending):
        super(SortOrder, self).__init__(child=child, direction=direction)


"""
strings
"""


class Ascii(_UnaryExpression):
    def __init__(self, child):
        super(Ascii, self).__init__(child=child)


class Base64(_UnaryExpression):
    def __init__(self, child):
        super(Base64, self).__init__(child=child)


class Concat(_WithChildren):
    def __init__(self, children=()):
        super(Concat, self).__init__(children=children)


@param('sep')
class ConcatWs(_WithChildren):
    def __init__(self, children=(), sep=''):
        super(ConcatWs, self).__init__(sep=sep, children=children)

    @staticmethod
    def _param_to_json(pc, value):
        if pc.name == 'sep':
            return to_json(value, pc.param_type)
        return _WithChildren._param_to_json(pc, value)


class Contains(_BinaryExpression):
    def __init__(self, left, right):
        super(Contains, self).__init__(left=left, right=right)


@param('charset')
class Decode(_UnaryExpression):
    def __init__(self, child, charset='ascii'):
        super(Decode, self).__init__(child=child, charset=charset)


@param('charset')
class Encode(_UnaryExpression):
    def __init__(self, child, charset='ascii'):
        super(Encode, self).__init__(child=child, charset=charset)


class EndsWith(_BinaryExpression):
    def __init__(self, left, right):
        super(EndsWith, self).__init__(left=left, right=right)


@param('precision', param_type='int')
class FormatNumber(_UnaryExpression):
    def __init__(self, child, precision=2):
        super(FormatNumber, self).__init__(child=child, precision=precision)


@param('fmt', server_name='format')
class FormatString(_WithChildren):
    def __init__(self, fmt='', children=()):
        super(FormatString, self).__init__(fmt=fmt, children=children)

    @staticmethod
    def _param_to_json(pc, value):
        if pc.name == 'fmt':
            return to_json(value, pc.param_type)
        return _WithChildren._param_to_json(pc, value)


class InitCap(_UnaryExpression):
    def __init__(self, child):
        super(InitCap, self).__init__(child=child)


@param('sub_str', server_name='subStr')
class StringInstr(_UnaryExpression):
    def __init__(self, child, sub_str=''):
        super(StringInstr, self).__init__(child=child, sub_str=sub_str)


class Length(_UnaryExpression):
    def __init__(self, child):
        super(Length, self).__init__(child=child)


@param('literal')
class Like(_UnaryExpression):
    def __init__(self, child, literal=''):
        super(Like, self).__init__(child=child, literal=literal)


class Lower(_UnaryExpression):
    def __init__(self, child):
        super(Lower, self).__init__(child=child)


class Levenshtein(_BinaryExpression):
    def __init__(self, left, right):
        super(Levenshtein, self).__init__(left=left, right=right)


@param('substr')
@param('pos', param_type='int')
class StringLocate(_UnaryExpression):
    def __init__(self, child, substr='', pos=0):
        super(StringLocate, self).__init__(child=child, substr=substr, pos=pos)


@param('length', param_type='int', server_name='len')
@param('padding', server_name='pad')
class StringLPad(_UnaryExpression):
    def __init__(self, child, length=0, padding=''):
        super(StringLPad, self).__init__(child=child, length=length, padding=padding)


class StringTrimLeft(_UnaryExpression):
    def __init__(self, child):
        super(StringTrimLeft, self).__init__(child=child)


@param('expr')
@param('group_idx', param_type='int', server_name='groupIdx')
class RegExpExtract(_UnaryExpression):
    def __init__(self, child, expr='', group_idx=0):
        super(RegExpExtract, self).__init__(child=child, expr=expr, group_idx=group_idx)


@param('pattern')
@param('replacement')
class RegExpReplace(_UnaryExpression):
    def __init__(self, child, pattern='', replacement=''):
        super(RegExpReplace, self).__init__(child=child, pattern=pattern, replacement=replacement)


@param('literal')
class RLike(_UnaryExpression):
    def __init__(self, child, literal=''):
        super(RLike, self).__init__(child=child, literal=literal)


class StartsWith(_BinaryExpression):
    def __init__(self, left, right):
        super(StartsWith, self).__init__(left=left, right=right)


@param('length', param_type='int', server_name='len')
@param('padding', server_name='pad')
class StringRPad(_UnaryExpression):
    def __init__(self, child, length=0, padding=''):
        super(StringRPad, self).__init__(child=child, length=length, padding=padding)


@param('n', param_type='int')
class StringRepeat(_UnaryExpression):
    def __init__(self, child, n=1):
        super(StringRepeat, self).__init__(child=child, n=n)


class StringReverse(_UnaryExpression):
    def __init__(self, child):
        super(StringReverse, self).__init__(child=child)


class StringTrimRight(_UnaryExpression):
    def __init__(self, child):
        super(StringTrimRight, self).__init__(child=child)


class SoundEx(_UnaryExpression):
    def __init__(self, child):
        super(SoundEx, self).__init__(child=child)


@param('pattern')
class StringSplit(_UnaryExpression):
    def __init__(self, child, pattern=''):
        super(StringSplit, self).__init__(child=child, pattern=pattern)


@param('matching')
@param('replace')
class StringTranslate(_UnaryExpression):
    def __init__(self, child, matching='', replace=''):
        super(StringTranslate, self).__init__(child=child, matching=matching, replace=replace)


class StringTrim(_UnaryExpression):
    def __init__(self, child):
        super(StringTrim, self).__init__(child=child)


@param('s', server_name='str')
@param('pos')
@param('length', server_name='len')
class Substring(Expression):
    def __init__(self, s, pos, length):
        super(Substring, self).__init__(s=s, pos=pos, length=length)


@param('delim')
@param('cnt', param_type='int', server_name='count')
class SubstringIndex(_UnaryExpression):
    def __init__(self, child, delim, cnt=1):
        super(SubstringIndex, self).__init__(child=child, delim=delim, cnt=cnt)


class UnBase64(_UnaryExpression):
    def __init__(self, child):
        super(UnBase64, self).__init__(child=child)


class Upper(_UnaryExpression):
    def __init__(self, child):
        super(Upper, self).__init__(child=child)


"""
Spark-specific
"""


@param('df_id', server_name='dfId')
@param('col_name', server_name='colName')
@server_namespace('io.cebes.spark.df.expressions')
class SparkPrimitiveExpression(Expression):
    def __init__(self, df_id='', col_name=''):
        super(SparkPrimitiveExpression, self).__init__(df_id=df_id, col_name=col_name)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six
import random
import pycebes.core.expressions as exprs
from pycebes.core.column import Column, lit


"""
This module contains functions that can be used to construct SQL expressions. 
Most of them take #Column as input and give #Column as output.
"""


def col(col_name='column'):
    """
    Returns a ``Column`` based on the given column name.
    """
    return Column(exprs.UnresolvedColumnName(col_name))


def _get_column(column):
    """
    Helper to get a ``Column`` object if the given argument is a string
    """
    if isinstance(column, six.text_type):
        column = col(column)
    if not isinstance(column, Column):
        raise ValueError('A Column or a column name (as a string) is expected. Got {!r}'.format(column))
    return column


def _with_expr(expr_class, cols, *args):
    """
    Helper to return a column constructed from the given expression class,
    with Column arguments in cols and additional arguments in args
    `cols` will be transformed into Column using :func:`_get_column`,
    other arguments in args will be passed as-it-is into the constructor of the expression class
    """
    if isinstance(cols, (tuple, list)):
        cols = tuple(_get_column(c).expr for c in cols)
    else:
        cols = _get_column(cols).expr
    return Column(expr_class(cols, *args))


def _with_binary_expr(expr_class, left, right, *args):
    """
    Helper to return a column constructed from the given binary expression class.
    """
    all_args = [_get_column(left).expr, _get_column(right).expr] + list(args)
    return Column(expr_class(*all_args))


"""
Aggregate functions
"""


def approx_count_distinct(column, rsd=0.05):
    """
    Returns the approximate number of distinct items in a group.

    # Arguments:
    column (Column): the column to compute
    rsd: maximum estimation error allowed (default = 0.05)
    """
    return _with_expr(exprs.ApproxCountDistinct, column, rsd)


def avg(column):
    """
    Returns the average of the values in a group
    """
    return _with_expr(exprs.Average, column)


def collect_list(column):
    """
    Returns a list of objects with duplicates
    """
    return _with_expr(exprs.CollectList, column)


def collect_set(column):
    """
    Returns a set of objects with duplicate elements eliminated
    """
    return _with_expr(exprs.CollectSet, column)


def corr(column1, column2):
    """
    Returns the Pearson Correlation Coefficient for two columns
    """
    return _with_binary_expr(exprs.Corr, left=column1, right=column2)


def count(column):
    """
    returns the number of items in a group
    """
    return _with_expr(exprs.Count, column)


def count_distinct(*args):
    """
    returns the number of distinct items in a group
    """
    cols = [_get_column(c).expr for c in args]
    return Column(exprs.CountDistinct(*cols))


def covar_pop(column1, column2):
    """
    Returns the population covariance for two columns
    """
    return _with_binary_expr(exprs.CovPopulation, left=column1, right=column2)


def covar_samp(column1, column2):
    """
    Returns the sample covariance for two columns
    """
    return _with_binary_expr(exprs.CovSample, left=column1, right=column2)


def first(column, ignore_nulls=False):
    """
    Returns the first value in a group.

    The function by default returns the first values it sees. It will return the first non-null
    value it sees when ``ignore_nulls`` is set to true. If all values are null, then null is returned.
    """
    return _with_expr(exprs.First, column, ignore_nulls)


def last(column, ignore_nulls=False):
    """
    Returns the last value in a group.

    The function by default returns the last values it sees. It will return the last non-null
    value it sees when ``ignore_nulls`` is set to true. If all values are null, then null is returned.
    """
    return _with_expr(exprs.Last, column, ignore_nulls)


def grouping(column):
    """
    indicates whether a specified column in a GROUP BY list is aggregated
    or not, returns 1 for aggregated or 0 for not aggregated in the result set.
    """
    return _with_expr(exprs.Grouping, column)


def grouping_id(*columns):
    """
    returns the level of grouping, equals to

    ``(grouping(c1) << (n-1)) + (grouping(c2) << (n-2)) + ... + grouping(cn)``

    > the list of columns should match with grouping columns exactly, or empty (means all the
    >    grouping columns).
    """
    return _with_expr(exprs.GroupingID, columns)


def kurtosis(column):
    """
    Returns the kurtosis of the values in a group
    """
    return _with_expr(exprs.Kurtosis, column)


def max(column):
    """
    Returns the kurtosis of the values in a group
    """
    return _with_expr(exprs.Max, column)


mean = avg


def min(column):
    """
    Returns the kurtosis of the values in a group
    """
    return _with_expr(exprs.Min, column)


def skewness(column):
    """
    Returns the kurtosis of the values in a group
    """
    return _with_expr(exprs.Skewness, column)


def stddev_samp(column):
    """
    Returns the sample standard deviation of the expression in a group.
    """
    return _with_expr(exprs.StddevSamp, column)


stddev = stddev_samp


def stddev_pop(column):
    """
    Returns the population standard deviation of the expression in a group.
    """
    return _with_expr(exprs.StddevPop, column)


def sum(column, is_distinct=False):
    """
    Returns the sum of all values in the expression

    # Arguments
    column (Column): the column to compute the sum
    is_distinct (bool): whether to only compute the sum of distinct values.
            False by default, meaning computing the sum of all values
    """
    return _with_expr(exprs.Sum, column, is_distinct)


def var_samp(column):
    """
    Returns the unbiased variance of the values in a group

    See #var_pop for computing the population variance.
    """
    return _with_expr(exprs.VarianceSamp, column)


variance = var_samp


def var_pop(column):
    """
    Returns the population variance of the values in a group

    See #var_samp for computing the unbiased variance.
    """
    return _with_expr(exprs.VariancePop, column)


"""
Non-aggregate functions
"""


def when(condition, value):
    """
    Evaluates a list of conditions and returns one of multiple possible result expressions.
    If otherwise is not defined at the end, null is returned for unmatched conditions.

    # Arguments
    condition (Column): the condition
    value: value to take when condition is true

    # Example
    ```python
    # encoding gender string column into integer
    people.select(functions.when(people.gender == 'male', 0)
        .when(people.gender == 'female', 1)
        .otherwise(2))
    ```
    """
    if not isinstance(condition, Column):
        raise ValueError('condition must be a Column expression')
    return Column(exprs.CaseWhen([(condition.expr, lit(value).expr)]))


def array(*columns):
    """
    Creates a new array column. The input columns must all have the same data type.
    """
    return _with_expr(exprs.CreateArray, columns)


def create_map(*columns):
    """
    Creates a new map column. The input columns must be grouped as key-value pairs, e.g.
    (key1, value1, key2, value2, ...). The key columns must all have the same data type, and can't
    be null. The value columns must all have the same data type.
    """
    return _with_expr(exprs.CreateMap, columns)


def coalesce(*columns):
    """
    Returns the first column that is not null, or null if all inputs are null.

    For example, ``coalesce(a, b, c)`` will return a if a is not null,
    or b if a is null and b is not null, or c if both a and b are null but c is not null.
    """
    return _with_expr(exprs.Coalesce, columns)


def input_file_name():
    """
    Spark-specific: Creates a string column for the file name of the current Spark task
    """
    return Column(exprs.InputFileName())


def is_nan(column):
    """
    Return true iff the column is NaN
    """
    return _with_expr(exprs.IsNaN, column)


def is_null(column):
    """
    Return true iff the column is null
    """
    return _with_expr(exprs.IsNull, column)


def monotonically_increasing_id():
    """
    Spark-specific: A column expression that generates monotonically increasing 64-bit integers.

    The generated ID is guaranteed to be monotonically increasing and unique, but not consecutive.
    The current implementation puts the partition ID in the upper 31 bits, and the record number
    within each partition in the lower 33 bits. The assumption is that the data frame has
    less than 1 billion partitions, and each partition has less than 8 billion records.

    As an example, consider a ``Dataframe`` with two partitions, each with 3 records.
    This expression would return the following IDs:

    0, 1, 2, 8589934592 (1L << 33), 8589934593, 8589934594.
    """
    return Column(exprs.MonotonicallyIncreasingID())


def nanvl(column1, column2):
    """
    Returns col1 if it is not NaN, or col2 if col1 is NaN.
    Both inputs should be floating point columns (DoubleType or FloatType).
    """
    return _with_binary_expr(exprs.NaNvl, left=column1, right=column2)


def negate(column):
    """
    Unary minus, i.e. negate the expression.

    # Example
    ```python
    # Select the amount column and negates all values.
    df.select(-df.amount)
    df.select(functions.negate(df.amount))
    df.select(functions.negate("amount"))
    ```
    """
    return _with_expr(exprs.UnaryMinus, column)


def rand(seed=None):
    """
    Generate a random column with i.i.d. samples from U[0.0, 1.0].
    ``seed`` will be automatically randomly generated by ``random.randint(0, 1000)`` if not specified.

    > This is un-deterministic when data partitions are not fixed.
    """
    if seed is None:
        seed = random.randint(0, 1000)
    return Column(exprs.Rand(seed))


def randn(seed=None):
    """
    Generate a column with i.i.d. samples from the standard normal distribution.
    ``seed`` will be automatically randomly generated by ``random.randint(0, 1000)`` if not specified.

    > This is un-deterministic when data partitions are not fixed.
    """
    if seed is None:
        seed = random.randint(0, 1000)
    return Column(exprs.Randn(seed))


def spark_partition_id():
    """
    Spark-specific: Partition ID of the Spark task.

    > This is un-deterministic because it depends on data partitioning and task scheduling.
    """
    return Column(exprs.SparkPartitionID())


def struct(*columns):
    """
    Creates a new struct column.
    If the input column is a column in a ``Dataframe``, or a derived column expression
    that is named (i.e. aliased), its name would be remained as the StructField's name,
    otherwise, the newly generated StructField's name would be auto generated as col${index + 1},
    i.e. col1, col2, col3, ...
    """
    return _with_expr(exprs.CreateStruct, cols=columns)


def expr(expr_str=''):
    """
    Parses the expression string into the column that it represents

    # Example
    ```python
    # get the number of words of each length
    df.group_by(functions.expr("length(word)")).count()
    ```
    """

    return Column(exprs.RawExpression(expr_str))


"""
Math functions
"""


def abs(column):
    """
    Computes the absolute value
    """
    return _with_expr(exprs.Abs, column)


def sqrt(column):
    """
    Computes the square root of the specified float value.
    """
    return _with_expr(exprs.Sqrt, column)


def bitwise_not(column):
    """
    Computes bitwise NOT.
    """
    return _with_expr(exprs.BitwiseNot, column)


def acos(column):
    """
    Computes the cosine inverse of the given value; the returned angle is in the range 0.0 through pi.
    """
    return _with_expr(exprs.Acos, column)


def asin(column):
    """
    Computes the sine inverse of the given value; the returned angle is in the range -pi/2 through pi/2.
    """
    return _with_expr(exprs.Asin, column)


def atan(column):
    """
    Computes the tangent inverse of the given value
    """
    return _with_expr(exprs.Atan, column)


def atan2(x, y):
    """
    Returns the angle theta from the conversion of rectangular coordinates (x, y) to polar coordinates (r, theta)

    # Arguments
    x: the x-component, can be a ``Column``, a string (column name), or a float value
    y: the y-component, can be a ``Column``, a string (column name), or a float value
    """
    if isinstance(x, float):
        x = lit(x)
    if isinstance(y, float):
        y = lit(y)
    return _with_binary_expr(exprs.Atan2, left=x, right=y)


def bin(column):
    """
    An expression that returns the string representation of the binary value of the given long column.
    For example, bin("12") returns "1100".
    """
    return _with_expr(exprs.Bin, column)


def cbrt(column):
    """
    Computes the cube-root of the given value
    """
    return _with_expr(exprs.Cbrt, column)


def ceil(column):
    """
    Computes the ceiling of the given value.
    """
    return _with_expr(exprs.Ceil, column)


def conv(column, from_base=10, to_base=2):
    """
    Convert a number in a string column from one base to another
    """
    return _with_expr(exprs.Conv, column, from_base, to_base)


def cos(column):
    """
    Computes the cosine of the given value
    """
    return _with_expr(exprs.Cos, column)


def cosh(column):
    """
    Computes the hyperbolic cosine of the given value
    """
    return _with_expr(exprs.Cosh, column)


def exp(column):
    """
    Computes the exponential of the given value
    """
    return _with_expr(exprs.Exp, column)


def expm1(column):
    """
    Computes the exponential of the given value minus one
    """
    return _with_expr(exprs.Expm1, column)


def factorial(column):
    """
    Computes the factorial of the given value
    """
    return _with_expr(exprs.Factorial, column)


def floor(column):
    """
    Computes the floor of the given value
    """
    return _with_expr(exprs.Floor, column)


def greatest(*columns):
    """
    Returns the greatest value of the list of values, skipping null values.
    This function takes at least 2 parameters. It will return null iff all parameters are null.
    """
    return _with_expr(exprs.Greatest, columns)


def hex(column):
    """
    Computes hex value of the given column
    """
    return _with_expr(exprs.Hex, column)


def unhex(column):
    """
    Inverse of hex. Interprets each pair of characters as a hexadecimal number
    and converts to the byte representation of number.
    """
    return _with_expr(exprs.Unhex, column)


def hypot(x, y):
    """
    Computes ``sqrt(x^2 + y^2)`` without intermediate overflow or underflow
    See #atan2 to compute the angle.

    # Arguments
    x: can be a ``Column``, a string (column name), or a double value
    y: can be a ``Column``, a string (column name), or a double value
    """
    if isinstance(x, float):
        x = lit(x)
    if isinstance(y, float):
        y = lit(y)
    return _with_binary_expr(exprs.Hypot, left=x, right=y)


def least(*columns):
    """
    Returns the least value of the list of values, skipping null values.
    This function takes at least 2 parameters. It will return null iff all parameters are null.
    """
    return _with_expr(exprs.Least, columns)


def log(column, base=None):
    """
    Computes the natural logarithm of the given value if ``base`` is None.
    If ``base`` is not None, compute the logarithm with the given base of the column.
    """
    if base is None:
        return _with_expr(exprs.Log, column)

    if not isinstance(base, float):
        raise ValueError('Expected a floating number base, got {!r}'.format(base))
    return _with_expr(exprs.Logarithm, column, base)


def log10(column):
    """
    Computes the logarithm of the given value in base 10
    """
    return _with_expr(exprs.Log10, column)


def log1p(column):
    """
    Computes the natural logarithm of the given value plus one
    """
    return _with_expr(exprs.Log1p, column)


def log2(column):
    """
    Computes the logarithm of the given column in base 2
    """
    return _with_expr(exprs.Log2, column)


def pow(left, right):
    """
    Returns the value of the first argument raised to the power of the second argument

    # Arguments
    left: can be a ``Column``, a string (column name), or a double value
    right: can be a ``Column``, a string (column name), or a double value
    """
    if isinstance(left, float):
        left = lit(left)
    if isinstance(right, float):
        right = lit(right)
    return _with_binary_expr(exprs.Pow, left=left, right=right)


def pmod(dividend, divisor):
    """
    Returns the positive value of dividend mod divisor.
    ``dividend`` and ``divisor`` can be ``Column`` or strings (column names)
    """
    return _with_binary_expr(exprs.Pmod, left=dividend, right=divisor)


def rint(column):
    """
    Returns the double value that is closest in value to the argument and
    is equal to a mathematical integer.
    """
    return _with_expr(exprs.Rint, column)


def round(column, scale=0):
    """
    Round the value of ``column`` to ``scale`` decimal places if ``scale >= 0``
    or at integral part when ``scale < 0``
    """
    return _with_binary_expr(exprs.Round, left=column, right=lit(scale))


def bround(column, scale=0):
    """
    Round the value of ``column`` to ``scale`` decimal places with HALF_EVEN round mode if ``scale >= 0``
    or at integral part when ``scale < 0``
    """
    return _with_binary_expr(exprs.BRound, left=column, right=lit(scale))


def shift_left(column, num_bits=1):
    """
    Shift the given value ``num_bits`` left. If the given value is a long value, this function
    will return a long value else it will return an integer value.
    """
    return _with_expr(exprs.ShiftLeft, column, num_bits)


def shift_right(column, num_bits=1):
    """
    Shift the given value ``num_bits`` right. If the given value is a long value, this function
    will return a long value else it will return an integer value.
    """
    return _with_expr(exprs.ShiftRight, column, num_bits)


def shift_right_unsigned(column, num_bits=1):
    """
    Unsigned shift the given value ``num_bits`` right. If the given value is a long value, this function
    will return a long value else it will return an integer value.
    """
    return _with_expr(exprs.ShiftRightUnsigned, column, num_bits)


def signum(column):
    """
    Computes the signum of the given value
    """
    return _with_expr(exprs.Signum, column)


def sin(column):
    """
    Computes the sine of the given value
    """
    return _with_expr(exprs.Sin, column)


def sinh(column):
    """
    Computes the hyperbolic sine of the given value
    """
    return _with_expr(exprs.Sinh, column)


def tan(column):
    """
    Computes the tangent of the given value
    """
    return _with_expr(exprs.Tan, column)


def tanh(column):
    """
    Computes the hyperbolic tangent of the given value
    """
    return _with_expr(exprs.Tanh, column)


def to_degrees(column):
    """
    Converts an angle measured in radians to an approximately equivalent angle measured in degrees
    """
    return _with_expr(exprs.ToDegrees, column)


def to_radians(column):
    """
    Converts an angle measured in degrees to an approximately equivalent angle measured in radians
    """
    return _with_expr(exprs.ToRadians, column)


"""
Misc functions
"""


def md5(column):
    """
    Calculates the MD5 digest of a binary column and returns the value
    as a 32 character hex string
    """
    return _with_expr(exprs.Md5, column)


def sha1(column):
    """
    Calculates the SHA-1 digest of a binary column and returns the value
    as a 40 character hex string
    """
    return _with_expr(exprs.Sha1, column)


def sha2(column, num_bits=0):
    """
    Calculates the SHA-2 family of hash functions of a binary column and
    returns the value as a hex string

    # Arguments
    column: column to compute SHA-2 on. Can be a ``Column`` or a string (a column name)
    num_bits: one of 224, 256, 384, or 512
    """
    if num_bits not in (0, 224, 256, 384, 512):
        raise ValueError('Invalid value of num_bits')
    return _with_expr(exprs.Sha2, column, num_bits)


def crc32(column):
    """
    Calculates the cyclic redundancy check value  (CRC32) of a binary column and
    returns the value as a bigint
    """
    return _with_expr(exprs.Crc32, column)


def hash(column):
    """
    Calculates the hash code of given columns, and returns the result as an int column
    """
    return _with_expr(exprs.Murmur3Hash, column)


"""
String functions
"""


def ascii(column):
    """
    Computes the numeric value of the first character of the string column, and returns the
    result as an int column
    """
    return _with_expr(exprs.Ascii, column)


def base64(column):
    """
    Computes the BASE64 encoding of a binary column and returns it as a string column.
    This is the reverse of unbase64
    """
    return _with_expr(exprs.Base64, column)


def concat(*columns):
    """
    Concatenates multiple input string columns together into a single string column
    """
    return _with_expr(exprs.Concat, columns)


def concat_ws(sep=' ', *columns):
    """
    Concatenates multiple input string columns together into a single string column,
    using the given separator
    """
    return _with_expr(exprs.ConcatWs, columns, sep)


def decode(column, charset='US-ASCII'):
    """
    Computes the first argument into a string from a binary using the provided character set
    (one of 'US-ASCII', 'ISO-8859-1', 'UTF-8', 'UTF-16BE', 'UTF-16LE', 'UTF-16').

    If either argument is null, the result will also be null.
    """
    return _with_expr(exprs.Decode, column, charset)


def encode(column, charset='US-ASCII'):
    """
    Computes the first argument into a binary from a string using the provided character set
    (one of 'US-ASCII', 'ISO-8859-1', 'UTF-8', 'UTF-16BE', 'UTF-16LE', 'UTF-16').

    If either argument is null, the result will also be null.
    """
    return _with_expr(exprs.Encode, column, charset)


def format_number(column, d=2):
    """
    Formats numeric column x to a format like '#,###,###.##', rounded to d decimal places,
    and returns the result as a string column.

        * If d is 0, the result has no decimal point or fractional part.
        * If d < 0, the result will be null.
    """
    return _with_expr(exprs.FormatNumber, column, d)


def format_string(fmt, *columns):
    """
    Formats the arguments in printf-style and returns the result as a string column
    """
    columns = (_get_column(c) for c in columns)
    return Column(exprs.FormatString(fmt, columns))


def initcap(column):
    """
    Returns a new string column by converting the first letter of each word to uppercase.
    Words are delimited by whitespace.
    For example, "hello world" will become "Hello World".
    """
    return _with_expr(exprs.InitCap, column)


def instr(column, substr=''):
    """
    Locate the position of the first occurrence of substr column in the given string.
    Returns null if either of the arguments are null.

    > The position is not zero based, but 1 based index, returns 0 if substr
    >    could not be found in ``column``
    """
    return _with_expr(exprs.StringInstr, column, substr)


def length(column):
    """
    Computes the length of a given string or binary column
    """
    return _with_expr(exprs.Length, column)


def lower(column):
    """
    Converts a string column to lower case
    """
    return _with_expr(exprs.Lower, column)


def levenshtein(left, right):
    """
    Computes the Levenshtein distance of the two given string columns
    """
    return _with_binary_expr(exprs.Levenshtein, left=left, right=right)


def locate(substr, column, pos=None):
    """
    Locate the position of the first occurrence of substr.
    If ``pos`` is not None, only search after position ``pos``.

    > The position is not zero based, but 1 based index, returns 0 if substr
    >    could not be found in str.
    """
    if not isinstance(substr, six.text_type):
        raise ValueError('Expected substr as a string, got {!r}'.format(substr))

    return _with_expr(exprs.StringLocate, column, substr, pos)


def lpad(column, pad_length=1, pad=''):
    """
    Left-pad the string column
    """
    return _with_expr(exprs.StringLPad, column, pad_length, pad)


def ltrim(column):
    """
    Trim the spaces from left end for the specified string value
    """
    return _with_expr(exprs.StringTrimLeft, column)


def regexp_extract(column, regexp='', group_idx=0):
    """
    Extract a specific group matched by a Java regex, from the specified string column.
    If the regex did not match, or the specified group did not match, an empty string is returned.
    """
    return _with_expr(exprs.RegExpExtract, column, regexp, group_idx)


def regexp_replace(column, pattern='', replacement=''):
    """
    Replace all substrings of the specified string value that match regexp with rep.
    """
    return _with_expr(exprs.RegExpReplace, column, pattern, replacement)


def unbase64(column):
    """
    Decodes a BASE64 encoded string column and returns it as a binary column.
    This is the reverse of base64.
    """
    return _with_expr(exprs.UnBase64, column)


def rpad(column, pad_length=1, pad=''):
    """
    Left-pad the string column
    """
    return _with_expr(exprs.StringRPad, column, pad_length, pad)


def repeat(column, n=1):
    """
    Repeats a string column n times, and returns it as a new string column
    """
    return _with_expr(exprs.StringRepeat, column, n)


def reverse(column):
    """
    Reverses the string column and returns it as a new string column
    """
    return _with_expr(exprs.StringReverse, column)


def rtrim(column):
    """
    Trim the spaces from right end for the specified string value
    """
    return _with_expr(exprs.StringTrimRight, column)


def soundex(column):
    """
    Return the soundex code for the specified expression
    """
    return _with_expr(exprs.SoundEx, column)


def split(column, pattern=''):
    """
    Splits str around pattern (pattern is a regular expression)

    > pattern is a string representation of the regular expression
    """
    return _with_expr(exprs.StringSplit, column, pattern)


def substring(column, pos=0, l=1):
    """
    Substring starts at ``pos`` and is of length ``l`` when str is String type or
    returns the slice of byte array that starts at ``pos`` in byte and is of length ``l``
    when ``column`` is Binary type

    # Arguments
    column: a ``Column`` object, or a column name
    pos: starting position, can be an integer, or a ``Column`` with its expression giving an integer value
    l: length of the substring, can be an integer, or a ``Column`` with its expression giving an integer value
    """
    return _with_expr(exprs.Substring, column, lit(pos).expr, lit(l).expr)


def substring_index(column, delim=' ', cnt=1):
    """
    Returns the substring from string ``column`` before ``cnt`` occurrences of the delimiter ``delim``.
    If ``cnt`` is positive, everything the left of the final delimiter (counting from left) is
    returned. If ``cnt`` is negative, every to the right of the final delimiter (counting from the
    right) is returned. substring_index performs a case-sensitive match when searching for ``delim``.
    """
    return _with_expr(exprs.SubstringIndex, column, delim, cnt)


def translate(column, matching_str='', replace_str=''):
    """
    Translate any character in the ``column`` by a character in ``replace_str``.
    The characters in ``replace_str`` correspond to the characters in ``matching_str``.
    The translate will happen when any character in the string matches the character
    in the ``matching_str``.
    """
    return _with_expr(exprs.StringTranslate, column, matching_str, replace_str)


def trim(column):
    """
    Trim the spaces from both ends for the specified string column
    """
    return _with_expr(exprs.StringTrim, column)


def upper(column):
    """
    Converts a string column to upper case
    """
    return _with_expr(exprs.Upper, column)


"""
DateTime functions
"""


def add_months(column, num_months=1):
    """
    Returns the date that is ``num_months`` after the date in ``column``
    """
    return _with_expr(exprs.AddMonths, column, num_months)


def current_date():
    """
    Returns the current date as a date column
    """
    return Column(exprs.CurrentDate())


def current_timestamp():
    """
    Returns the current timestamp as a timestamp column
    """
    return Column(exprs.CurrentTimestamp())


def date_format(column, fmt=''):
    """
    Converts a date/timestamp/string to a value of string in the format specified by the date
    format given by the second argument.

    A pattern could be for instance `dd.MM.yyyy` and could return a string like '18.03.1993'. All
    pattern letters of [[java.text.SimpleDateFormat]] can be used.

    > Use when ever possible specialized functions like :func:`year`. These benefit from a
    >     specialized implementation.
    """
    return _with_expr(exprs.DateFormatClass, column, fmt)


def date_add(column, days=1):
    """
    Returns the date that is `days` days after the date in ``column``
    """
    return _with_expr(exprs.DateAdd, column, days)


def date_sub(column, days=1):
    """
    Returns the date that is `days` days before the date in ``column``
    """
    return _with_expr(exprs.DateSub, column, days)


def datediff(end_column, start_column):
    """
    Returns the number of days from `start` to `end`
    """
    return _with_binary_expr(exprs.DateDiff, end_column, start_column)


def year(column):
    """
    Extracts the year as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.Year, column)


def quarter(column):
    """
    Extracts the quarter as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.Quarter, column)


def month(column):
    """
    Extracts the month as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.Month, column)


def dayofmonth(column):
    """
    Extracts the day of the month as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.DayOfMonth, column)


def dayofyear(column):
    """
    Extracts the day of the year as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.DayOfYear, column)


def hour(column):
    """
    Extracts the hours as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.Hour, column)


def last_day(column):
    """
    Given a date column, returns the last day of the month which the given date belongs to.
    For example, input "2015-07-27" returns "2015-07-31" since July 31 is the last day of the
    month in July 2015.
    """
    return _with_expr(exprs.LastDay, column)


def minute(column):
    """
    Extracts the minutes as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.Minute, column)


def months_between(column1, column2):
    """
    Returns number of months between dates `column1` and `column2`
    """
    return _with_binary_expr(exprs.MonthsBetween, column1, column2)


def next_day(column, day_of_week='mon'):
    """
    Given a date column, returns the first date which is later than the value of the date column
    that is on the specified day of the week.

    For example, `next_day('2015-07-27', "Sunday")` returns 2015-08-02 because that is the first
    Sunday after 2015-07-27.

    Day of the week parameter is case insensitive, and accepts:

        "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun".
    """
    return _with_expr(exprs.NextDay, column, day_of_week)


def second(column):
    """
    Extracts the seconds as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.Second, column)


def weekofyear(column):
    """
    Extracts the week number as an integer from a given date/timestamp/string
    """
    return _with_expr(exprs.WeekOfYear, column)


def from_unixtime(column, fmt='yyyy-MM-dd HH:mm:ss'):
    """
    Converts the number of seconds from unix epoch (1970-01-01 00:00:00 UTC) to a string
    representing the timestamp of that moment in the current system time zone in the given
    format.
    """
    return _with_expr(exprs.FromUnixTime, column, fmt)


def unix_timestamp(column=None, pattern='yyyy-MM-dd HH:mm:ss'):
    """
    Convert time string in ``column`` with given ``pattern``
    (see [http://docs.oracle.com/javase/tutorial/i18n/format/simpleDateFormat.html])
    to Unix time stamp (in seconds), return null if fail.

    If ``column=None``, the current Unix timestamp (computed by :func:`current_timestamp`) will be used
    """
    if column is None:
        column = current_timestamp()
    return _with_expr(exprs.UnixTimestamp, column, pattern)


def to_date(column):
    """
    Converts the column into DateType
    """
    return _with_expr(exprs.ToDate, column)


def trunc(column, fmt='year'):
    """
    Returns date truncated to the unit specified by the format.

    # Arguments
    column: the source column, can be a ``Column`` object or a string (column name)
    fmt: 'year', 'yyyy', 'yy' for truncate by year,
            or 'month', 'mon', 'mm' for truncate by month
    """
    return _with_expr(exprs.TruncDate, column, fmt)


def from_utc_timestamp(column, tz=''):
    """
    Assumes given timestamp is UTC and converts to given timezone
    """
    return _with_expr(exprs.FromUTCTimestamp, column, tz)


def to_utc_timestamp(column, tz=''):
    """
    Assumes given timestamp is in given timezone and converts to UTC
    """
    return _with_expr(exprs.ToUTCTimestamp, column, tz)


def window(column, window_duration='', slide_duration=None, start_time='0 second'):
    """
    Bucketize rows into one or more time windows given a timestamp specifying column. Window
    starts are inclusive but the window ends are exclusive, e.g. ``12:05`` will be in the window
    ``[12:05,12:10)`` but not in ``[12:00,12:05)``.

    Windows can support microsecond precision. Windows in the order of months are not supported.

    The following example takes the average stock price for a one minute window every 10 seconds
    starting 5 seconds after the hour:

    # Example
    ```python
    df = ...  # schema: timestamp: TimestampType, stockId: StringType, price: DoubleType
    df.group_by(functions.window(df.timestamp, "1 minute", "10 seconds", "5 seconds"), df.stockId)
        .agg(mean("price"))
    ```

    The windows will look like:

    ```python
        09:00:05-09:01:05
        09:00:15-09:01:15
        09:00:25-09:01:25 ...
    ```

    For a streaming query, you may use the function ``current_timestamp`` to generate windows on
    processing time.

    # Arguments
    column: The column or the expression to use as the timestamp for windowing by time.
                   The time column must be of TimestampType.
    window_duration: A string specifying the width of the window, e.g. `10 minutes`,
                    `1 second`. Check `CalendarIntervalType` for
                    valid duration identifiers. Note that the duration is a fixed length of
                    time, and does not vary over time according to a calendar. For example,
                    `1 day` always means `86,400,000 milliseconds`, not a calendar day.
    slide_duration: A string specifying the sliding interval of the window, e.g. `1 minute`.
                    A new window will be generated every `slide_duration`. Must be less than
                    or equal to the `window_duration`.
                    This duration is likewise absolute, and does not vary according to a calendar.
                    If unspecified, `slide_duration` will be equal to `window_duration`
    start_time (str): The offset with respect to `1970-01-01 00:00:00 UTC` with which to start
                        window intervals. For example, in order to have hourly tumbling windows
                        that start 15 minutes past the hour, e.g. `12:15 - 13:15`, `13:15 - 14:15`...
                        provide `start_time` as `15 minutes`.
    """
    if slide_duration is None:
        slide_duration = window_duration
    return _with_expr(exprs.TimeWindow, column, window_duration, slide_duration, start_time)


"""
Collection functions
"""


def array_contains(column, value):
    """
    Returns true if the array contains ``value``
    """
    return _with_binary_expr(exprs.ArrayContains, left=column, right=lit(value))


def explode(column):
    """
    Creates a new row for each element in the given array or map column
    """
    return _with_expr(exprs.Explode, column)


def posexplode(column):
    """
    Creates a new row for each element with position in the given array or map column
    """
    return _with_expr(exprs.PosExplode, column)


def get_json_object(column, path=''):
    """
    Extracts json object from a json string based on json path specified, and returns json string
    of the extracted json object. It will return null if the input json string is invalid.
    """
    return _with_expr(exprs.GetJsonObject, column, path)


def json_tuple(column, *fields):
    """
    Creates a new row for a json column according to the given field names
    """
    if len(fields) == 0:
        raise ValueError('at least 1 field name should be given')
    return _with_expr(exprs.JsonTuple, column, fields)


def size(column):
    """
    Returns length of array or map
    """
    return _with_expr(exprs.Size, column)


def sort_array(column, asc=True):
    """
    Sorts the input array for the given column in ascending / descending order,
    according to the natural ordering of the array elements.
    """
    return _with_expr(exprs.SortArray, column, asc)
